import com.imsl.datamining.neural.*;
import com.imsl.math.PrintMatrix;
import com.imsl.math.PrintMatrixFormat;
import java.io.*;
import java.util.logging.*;

//*****************************************************************************
// Three Layer Feed-Forward Network with 4 inputs, all 
// continuous, and 3 classification categories.
//
//  new classification training_ex5.c
//
// This is perhaps the best known database to be found in the pattern
//     recognition literature.  Fisher's paper is a classic in the field.
//     The data set contains 3 classes of 50 instances each,
//     where each class refers to a type of iris plant.  One class is
//     linearly separable from the other 2; the latter are NOT linearly
//     separable from each other.
//
//  Predicted attribute: class of iris plant.
//     1=Iris Setosa, 2=Iris Versicolour, and 3=Iris Virginica
//
//  Input Attributes (4 Continuous Attributes)
//     X1: Sepal length, X2: Sepal width, X3: Petal length, and X4: Petal width
//*****************************************************************************

public class MultiClassificationEx1 implements Serializable {
    private static int nObs          = 150; // number of training patterns
    private static int nInputs       =  4; // 9 nominal coded as 0=x, 1=o, 2=blank
    private static int nOutputs      =  3; // one continuous output (nClasses=2)
    private static boolean trace     = true; // Turns on/off training log

    
	// irisData[]:  The raw data matrix.  This is a 2-D matrix with 150 rows and 5 columns.  *
	//              The first 4 columns are the continuous input attributes and the 5th      *
	//              column is the classification category (1-3).  These data contain no      *
	//              categorical input attributes.                                            *

    private static double[][] irisData = {
        {5.1,3.5,1.4,0.2,1},{4.9,3.0,1.4,0.2,1},{4.7,3.2,1.3,0.2,1},{4.6,3.1,1.5,0.2,1},
        {5.0,3.6,1.4,0.2,1},{5.4,3.9,1.7,0.4,1},{4.6,3.4,1.4,0.3,1},{5.0,3.4,1.5,0.2,1},
        {4.4,2.9,1.4,0.2,1},{4.9,3.1,1.5,0.1,1},{5.4,3.7,1.5,0.2,1},{4.8,3.4,1.6,0.2,1},
        {4.8,3.0,1.4,0.1,1},{4.3,3.0,1.1,0.1,1},{5.8,4.0,1.2,0.2,1},{5.7,4.4,1.5,0.4,1},
        {5.4,3.9,1.3,0.4,1},{5.1,3.5,1.4,0.3,1},{5.7,3.8,1.7,0.3,1},{5.1,3.8,1.5,0.3,1},
        {5.4,3.4,1.7,0.2,1},{5.1,3.7,1.5,0.4,1},{4.6,3.6,1.0,0.2,1},{5.1,3.3,1.7,0.5,1},
        {4.8,3.4,1.9,0.2,1},{5.0,3.0,1.6,0.2,1},{5.0,3.4,1.6,0.4,1},{5.2,3.5,1.5,0.2,1},
        {5.2,3.4,1.4,0.2,1},{4.7,3.2,1.6,0.2,1},{4.8,3.1,1.6,0.2,1},{5.4,3.4,1.5,0.4,1},
        {5.2,4.1,1.5,0.1,1},{5.5,4.2,1.4,0.2,1},{4.9,3.1,1.5,0.1,1},{5.0,3.2,1.2,0.2,1},
        {5.5,3.5,1.3,0.2,1},{4.9,3.1,1.5,0.1,1},{4.4,3.0,1.3,0.2,1},{5.1,3.4,1.5,0.2,1},
        {5.0,3.5,1.3,0.3,1},{4.5,2.3,1.3,0.3,1},{4.4,3.2,1.3,0.2,1},{5.0,3.5,1.6,0.6,1},
        {5.1,3.8,1.9,0.4,1},{4.8,3.0,1.4,0.3,1},{5.1,3.8,1.6,0.2,1},{4.6,3.2,1.4,0.2,1},
        {5.3,3.7,1.5,0.2,1},{5.0,3.3,1.4,0.2,1},

        {7.0,3.2,4.7,1.4,2},{6.4,3.2,4.5,1.5,2},{6.9,3.1,4.9,1.5,2},{5.5,2.3,4.0,1.3,2},
        {6.5,2.8,4.6,1.5,2},{5.7,2.8,4.5,1.3,2},{6.3,3.3,4.7,1.6,2},{4.9,2.4,3.3,1.0,2},
        {6.6,2.9,4.6,1.3,2},{5.2,2.7,3.9,1.4,2},{5.0,2.0,3.5,1.0,2},{5.9,3.0,4.2,1.5,2},
        {6.0,2.2,4.0,1.0,2},{6.1,2.9,4.7,1.4,2},{5.6,2.9,3.6,1.3,2},{6.7,3.1,4.4,1.4,2},
        {5.6,3.0,4.5,1.5,2},{5.8,2.7,4.1,1.0,2},{6.2,2.2,4.5,1.5,2},{5.6,2.5,3.9,1.1,2},
        {5.9,3.2,4.8,1.8,2},{6.1,2.8,4.0,1.3,2},{6.3,2.5,4.9,1.5,2},{6.1,2.8,4.7,1.2,2},
        {6.4,2.9,4.3,1.3,2},{6.6,3.0,4.4,1.4,2},{6.8,2.8,4.8,1.4,2},{6.7,3.0,5.0,1.7,2},
        {6.0,2.9,4.5,1.5,2},{5.7,2.6,3.5,1.0,2},{5.5,2.4,3.8,1.1,2},{5.5,2.4,3.7,1.0,2},
        {5.8,2.7,3.9,1.2,2},{6.0,2.7,5.1,1.6,2},{5.4,3.0,4.5,1.5,2},{6.0,3.4,4.5,1.6,2},
        {6.7,3.1,4.7,1.5,2},{6.3,2.3,4.4,1.3,2},{5.6,3.0,4.1,1.3,2},{5.5,2.5,4.0,1.3,2},
        {5.5,2.6,4.4,1.2,2},{6.1,3.0,4.6,1.4,2},{5.8,2.6,4.0,1.2,2},{5.0,2.3,3.3,1.0,2},
        {5.6,2.7,4.2,1.3,2},{5.7,3.0,4.2,1.2,2},{5.7,2.9,4.2,1.3,2},{6.2,2.9,4.3,1.3,2},
        {5.1,2.5,3.0,1.1,2},{5.7,2.8,4.1,1.3,2},

        {6.3,3.3,6.0,2.5,3},{5.8,2.7,5.1,1.9,3},{7.1,3.0,5.9,2.1,3},{6.3,2.9,5.6,1.8,3},
        {6.5,3.0,5.8,2.2,3},{7.6,3.0,6.6,2.1,3},{4.9,2.5,4.5,1.7,3},{7.3,2.9,6.3,1.8,3},
        {6.7,2.5,5.8,1.8,3},{7.2,3.6,6.1,2.5,3},{6.5,3.2,5.1,2.0,3},{6.4,2.7,5.3,1.9,3},
        {6.8,3.0,5.5,2.1,3},{5.7,2.5,5.0,2.0,3},{5.8,2.8,5.1,2.4,3},{6.4,3.2,5.3,2.3,3},
        {6.5,3.0,5.5,1.8,3},{7.7,3.8,6.7,2.2,3},{7.7,2.6,6.9,2.3,3},{6.0,2.2,5.0,1.5,3},
        {6.9,3.2,5.7,2.3,3},{5.6,2.8,4.9,2.0,3},{7.7,2.8,6.7,2.0,3},{6.3,2.7,4.9,1.8,3},
        {6.7,3.3,5.7,2.1,3},{7.2,3.2,6.0,1.8,3},{6.2,2.8,4.8,1.8,3},{6.1,3.0,4.9,1.8,3},
        {6.4,2.8,5.6,2.1,3},{7.2,3.0,5.8,1.6,3},{7.4,2.8,6.1,1.9,3},{7.9,3.8,6.4,2.0,3},
        {6.4,2.8,5.6,2.2,3},{6.3,2.8,5.1,1.5,3},{6.1,2.6,5.6,1.4,3},{7.7,3.0,6.1,2.3,3},
        {6.3,3.4,5.6,2.4,3},{6.4,3.1,5.5,1.8,3},{6.0,3.0,4.8,1.8,3},{6.9,3.1,5.4,2.1,3},
        {6.7,3.1,5.6,2.4,3},{6.9,3.1,5.1,2.3,3},{5.8,2.7,5.1,1.9,3},{6.8,3.2,5.9,2.3,3},
        {6.7,3.3,5.7,2.5,3},{6.7,3.0,5.2,2.3,3},{6.3,2.5,5.0,1.9,3},{6.5,3.0,5.2,2.0,3},
        {6.2,3.4,5.4,2.3,3},{5.9,3.0,5.1,1.8,3}
    };    
    
    public static void main(String[] args) throws Exception {
        double xData[][] = new double[nObs][nInputs];
        int yData[] = new int[nObs];
        
        for (int i = 0;  i < nObs;  i++) {
            for (int j = 0;  j < nInputs;  j++) {
                xData[i][j] = irisData[i][j];
            }
            yData[i] = (int)irisData[i][4];
        }
        
        // Create network
        FeedForwardNetwork network = new FeedForwardNetwork();
        network.getInputLayer().createInputs(nInputs);
        network.createHiddenLayer().createPerceptrons(4, Activation.LOGISTIC, 0.0);
        network.createHiddenLayer().createPerceptrons(3, Activation.LOGISTIC, 0.0);
        network.getOutputLayer().createPerceptrons(nOutputs, Activation.SOFTMAX, 0.0);
        network.linkAll();
        
        MultiClassification classification = new MultiClassification(network);
        
        // Create trainer
        QuasiNewtonTrainer trainer = new QuasiNewtonTrainer();
        trainer.setError(classification.getError());
        trainer.setMaximumTrainingIterations(1000);
        
        // If tracing is requested setup training logger
        if (trace) {
            Handler handler = new FileHandler("ClassificationNetworkTraining.log");
            Logger logger = Logger.getLogger("com.imsl.datamining.neural");
            logger.setLevel(Level.FINEST);
            logger.addHandler(handler);
            handler.setFormatter(QuasiNewtonTrainer.getFormatter());
        }
        // Train Network
        long t0 = System.currentTimeMillis();
        classification.train(trainer, xData, yData);
        
        // Display Network Errors
       double stats[] = classification.computeStatistics(xData, yData);
       System.out.println("***********************************************");
       System.out.println("--> Cross-entropy error:        "+(float)stats[0]);
       System.out.println("--> Classification error rate:  "+(float)stats[1]);
       System.out.println("***********************************************");
       System.out.println("");
        
        double weight[]   = network.getWeights();
        double gradient[] = trainer.getErrorGradient();
        double wg[][] = new double[weight.length][2];
        for(int i = 0;  i < weight.length;  i++) 
        {
            wg[i][0] = weight[i];
            wg[i][1] = gradient[i];
        }
        PrintMatrixFormat pmf = new PrintMatrixFormat();
        pmf.setNumberFormat(new java.text.DecimalFormat("0.000000"));
        pmf.setColumnLabels(new String[]{"Weights", "Gradients"});
        new PrintMatrix().print(pmf,wg);
        
        double report[][] = new double[nObs][nInputs+2];
        for (int i = 0;  i < nObs;  i++) {
            for (int j = 0;  j < nInputs;  j++) {
                report[i][j] = xData[i][j];
            }
            report[i][nInputs] = irisData[i][4];
            report[i][nInputs+1] = classification.predictedClass(xData[i]);
        }
        pmf = new PrintMatrixFormat();
        pmf.setColumnLabels(new String[]{
            "Sepal Length", 
            "Sepal Width",
            "Petal Length",
            "Petal Width",
            "Expected",
            "Predicted"});
        new PrintMatrix("Forecast").print(pmf, report);


    // **********************************************************************
    // DISPLAY CLASSIFICATION STATISTICS
    // **********************************************************************
       double statsClass[] = classification.computeStatistics(xData, yData);
       // Display Network Errors
       System.out.println("***********************************************");
       System.out.println("--> Cross-Entropy Error:      "+(float)statsClass[0]);
       System.out.println("--> Classification Error:     "+(float)statsClass[1]);
       System.out.println("***********************************************");
       System.out.println("");
        long t1 = System.currentTimeMillis();
        double small = 1.e-7;
        double time =  t1-t0; //Math.max(small, (double)(t1-t0)/(double)iters);
        time = time/1000;
        System.out.println("****************Time:  "+time);

        System.out.println("Cross-Entropy Error Value = "+trainer.getErrorValue());

    }
}
